/* Copyright 2002-2023 CS GROUP
 * Licensed to CS GROUP (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.tutorials.maneuvers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.SortedSet;
import java.util.TreeSet;

import org.hipparchus.Field;
import org.hipparchus.analysis.differentiation.DSFactory;
import org.hipparchus.analysis.differentiation.DerivativeStructure;
import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.ode.nonstiff.AdaptiveStepsizeFieldIntegrator;
import org.hipparchus.ode.nonstiff.AdaptiveStepsizeIntegrator;
import org.hipparchus.ode.nonstiff.DormandPrince853FieldIntegrator;
import org.hipparchus.ode.nonstiff.DormandPrince853Integrator;
import org.hipparchus.util.FastMath;
import org.hipparchus.util.MathUtils;
import org.orekit.attitudes.Attitude;
import org.orekit.attitudes.LofOffset;
import org.orekit.bodies.CelestialBodyFactory;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.forces.gravity.HolmesFeatherstoneAttractionModel;
import org.orekit.forces.gravity.ThirdBodyAttraction;
import org.orekit.forces.gravity.potential.GravityFieldFactory;
import org.orekit.forces.gravity.potential.NormalizedSphericalHarmonicsProvider;
import org.orekit.forces.gravity.potential.UnnormalizedSphericalHarmonicsProvider;
import org.orekit.forces.maneuvers.ImpulseManeuver;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.LOFType;
import org.orekit.frames.Transform;
import org.orekit.orbits.CartesianOrbit;
import org.orekit.orbits.CircularOrbit;
import org.orekit.orbits.EquinoctialOrbit;
import org.orekit.orbits.FieldEquinoctialOrbit;
import org.orekit.orbits.FieldOrbit;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.orbits.Orbit;
import org.orekit.orbits.PositionAngleType;
import org.orekit.propagation.FieldSpacecraftState;
import org.orekit.propagation.PropagationType;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.events.DateDetector;
import org.orekit.propagation.numerical.NumericalPropagator;
import org.orekit.propagation.sampling.OrekitFixedStepHandler;
import org.orekit.propagation.semianalytical.dsst.DSSTPropagator;
import org.orekit.propagation.semianalytical.dsst.FieldDSSTPropagator;
import org.orekit.propagation.semianalytical.dsst.forces.DSSTForceModel;
import org.orekit.propagation.semianalytical.dsst.forces.DSSTTesseral;
import org.orekit.propagation.semianalytical.dsst.forces.DSSTThirdBody;
import org.orekit.propagation.semianalytical.dsst.forces.DSSTZonal;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.ChronologicalComparator;
import org.orekit.time.FieldAbsoluteDate;
import org.orekit.time.TimeScale;
import org.orekit.time.TimeScalesFactory;
import org.orekit.tutorials.yaml.TutorialOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialCartesianOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialCircularOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialEquinoctialOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialKeplerianOrbit;
import org.orekit.utils.Constants;
import org.orekit.utils.PVCoordinates;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

/**
 * This tutorial shows the use of the FieldDSSTPropagator to calculate
 * an East/West station-keeping strategy in GEO.
 * <p>
 * The strategy aims to center a parabolic excursion in longitude around
 * a given target during a control cycle.
 * </p>
 * <p>
 * The maneuvers are calculated to bring the end-of-cycle longitude to a value
 * that will start a perfectly centered parabola with the first maneuver of the
 * next cycle. This works regardless of the initial point, which may be either
 * already in the longitude slot, west of the slot, or east of the slot.
 * </p>
 * <p>
 * If the starting point is too far from the station-keeping window, multiple
 * maneuvers are performed to reach the slot more quickly, up to the maximum
 * number of maneuvers allowed per cycle and within the dV size constraints.
 * </p>
 * @author CS Group
 */
public class StationKeeping {

    /** Earth gravitational constant used everywhere. */
    private static final double MU = Constants.IERS2010_EARTH_MU;

    /** Separator for csv file. */
    private static final String SEP = ", ";

    /** Earth frame to compute longitude. */
    private final Frame ecef;

    /** Target longitude. */
    private final double center;

    /** DV max (m/s). */
    private final double dVMax;

    /** ISP (s). */
    private final double isp;

    /** Maximum number of maneuvers per cycle. */
    private final int maxNbMan;

    /** Number of orbits between maneuvers. */
    private final int orbitsSeparation;

    /** Initial offset with respect to synchronous semi major axis. */
    private double a0Mas;

    /** Semi-major axis slope. */
    private double aDot;

    /** Initial longitude. */
    private double l0;

    /** dlDot/da. */
    private double dlDotDa;

    /** Constructor.
     * @param westBound the western boundary of the longitude window
     * @param eastBound the eastern boundary of the longitude window
     * @param dVMax     the maximum deltaV for each maneuver
     * @param isp       the ISP of the propulsion system
     * @param maxNbMan  the maximum number of maneuvers per cycle
     * @param orbitsNb  the number of orbits between maneuvers
     */
    public StationKeeping(final double westBound, final double eastBound,
                          final double dVMax, final double isp,
                          final int maxNbMan, final int orbitsNb) {
        // Set current parameters
        this.ecef = CelestialBodyFactory.getEarth().getBodyOrientedFrame();

        this.center = 0.5 * (westBound + MathUtils.normalizeAngle(eastBound, westBound));

        this.dVMax  = dVMax;
        this.isp    = isp;

        this.maxNbMan         = maxNbMan;
        this.orbitsSeparation = orbitsNb;
    }

    /**
     * Run the SK simulation for a single cycle.
     * <p>
     * The longitude drift &zeta; (i.e. time derivative of the mean longitude) depends
     * linearly on semi major axis, and is null at geosynchronous semi-major axis:
     * <pre>
     *  <code>dl(t)/dt = &zeta; = &part;&zeta;/&part;a &times; [a(t) - a<sub>s</sub>]</code>
     * </pre>
     * where a<sub>s</sub> is the synchronous semi-major axis and &part;&zeta;/&part;a
     * is a negative coefficient depending only on the gravity field with a theoretical
     * value for Keplerian motion &part;&zeta;/&part;a = -(3/2)&radic;(&mu;/a<sub>s</sub>).
     * <p>
     * The model for semi-major axis drift is limited to secular terms only here:
     * <pre>
     * <code>a(t) = a(t<sub>0</sub>) + &aring; &times; (t - t<sub>0</sub>)</code>
     * </pre>
     * where &aring; is the secular time derivative of the semi-major axis.
     * <p>
     * This implies the model for mean longitude is quadratic:
     * <pre>
     * <code>l(t) = l(t<sub>0</sub>) + &part;&zeta;/&part;a &times; [a(t<sub>0</sub>) - a<sub>s</sub>] &times; (t - t<sub>0</sub>) + &frac12; &part;&zeta;/&part;a &times; &aring; &times; (t - t<sub>0</sub>)<sup>2</sup></code>
     * </pre>
     * <p>
     * The initial longitude l(t<sub>0</sub>) is inherited from the previous
     * cycle motion. The curvature comes from &aring; which depends on the station-keeping
     * longitude slot. So the only control parameter left is the initial drift, which is
     * adjusted by changing the offset [a(t<sub>0</sub>) - a<sub>s</sub>] at initial time
     * t<sub>0</sub> thanks to tangential in-plane maneuvers. The aim of the control law is
     * to achieve a parabolic motion that is centered in the station-keeping window during
     * one cycle duration, i.e. with a peak (depending on acceleration direction, of course)
     * longitude at cycle middle time.
     * </p>
     * <p>
     * The model parameters &part;&zeta;/&part;a, a<sub>s</sub> and &aring; are
     * computed with the FieldDSSTPropagator used to propagate the mean orbital
     * elements and the partial derivatives.
     * </p>
     * @param initialState  the initial state at cycle start
     * @param cycleDuration the duration of the control cycle
     * @return the computed maneuvers
     */
    public List<ImpulseManeuver> run(final SpacecraftState initialState,
                                                   final double cycleDuration) {

        // Initializes the list of maneuvers
        final List<ImpulseManeuver> maneuvers = new ArrayList<>();

        // Computes the parameters of the mean models
        meanModel(initialState, cycleDuration);

        // compute the in-plane maneuver required to get the  required sma offset
        final double totalDV = totalDV(initialState, cycleDuration);

        // Computes the number of maneuvers required for SK
        // taking maneuvering capabilities into account
        final double limitDV = (totalDV < 0) ? -dVMax : dVMax;
        final int nbMan      = FastMath.min(maxNbMan, (int) FastMath.ceil(FastMath.abs(totalDV / limitDV)));
        final double dVMan   = FastMath.max(-dVMax, FastMath.min(dVMax, totalDV / nbMan));

        // in order to avoid tempering eccentricity too much,
        // we use a (n + 0.5) orbits between maneuvers
        final double separation = (orbitsSeparation + 0.5) * initialState.getKeplerianPeriod();

        // Creates all the required tangential maneuvers
        for (int i = 0; i < nbMan; i++) {
            final AbsoluteDate manDate = initialState.getDate().shiftedBy((i + 1) * separation);
            // Adds a tangential maneuver to the list
            maneuvers.add(new ImpulseManeuver(new DateDetector(manDate),
                                                new Vector3D(dVMan, Vector3D.PLUS_I), isp));
        }

        // Returns the maneuvers for the current cycle
        return maneuvers;
    }

    /** Get Earth based mean longitude.
     * @param orbit current orbit
     * @return Earth based mean longitude
     */
    public double getMeanLongitude(final Orbit orbit) {
        // compute sidereal time
        final Transform transform = ecef.getTransformTo(orbit.getFrame(), orbit.getDate());
        final double theta = transform.transformVector(Vector3D.PLUS_I).getAlpha();
        // returns Earth based mean longitude
        return MathUtils.normalizeAngle(orbit.getLM() - theta, center);
    }

    /**
     * Computes the parameters of the mean models.
     * @param initialState the initial state
     * @param cycleDuration the cycle duration
     */
    private void meanModel(final SpacecraftState initialState,
                           final double cycleDuration) {

        // Gets the orbit from the state
        final Orbit orbIni = initialState.getOrbit();

        // Factory for the DerivativeStructures
        final DSFactory factory = new DSFactory(2, 1);

        // Orbital parameters
        final DerivativeStructure a_0  = factory.variable(0, orbIni.getA());
        final DerivativeStructure ex_0 = factory.constant(orbIni.getEquinoctialEx());
        final DerivativeStructure ey_0 = factory.constant(orbIni.getEquinoctialEy());
        final DerivativeStructure hx_0 = factory.constant(orbIni.getHx());
        final DerivativeStructure hy_0 = factory.constant(orbIni.getHy());
        final DerivativeStructure lm_0 = factory.variable(1, orbIni.getLM());
        final DerivativeStructure mu   = factory.constant(orbIni.getMu());

        // Gets the field of the DerivativeStructure to build new instances
        final Field<DerivativeStructure> field = a_0.getField();

        // Builds the field date
        final FieldAbsoluteDate<DerivativeStructure> t_0 = new FieldAbsoluteDate<>(field, orbIni.getDate());

        // Creates the field orbit
        final FieldOrbit<DerivativeStructure> fieldOrbIni = new FieldEquinoctialOrbit<>(
                a_0, ex_0, ey_0, hx_0, hy_0, lm_0, PositionAngleType.MEAN, orbIni.getFrame(), t_0, mu);

        // Creates the field propagator
        final FieldDSSTPropagator<DerivativeStructure> propagator = createFieldPropagator(fieldOrbIni);

        // Propagate over the cycle
        final FieldSpacecraftState<DerivativeStructure> finalState = propagator.propagate(t_0.shiftedBy(cycleDuration));

        // Extract model data such as:
        // dl(t)/dt = dlDotDa * [a(t) - as]
        // a(t)     = a(t0) + aDot * (t - t0)
        // l(t)     = l(t0) + dlDotDa * [a(t0) - as] * (t - t0) + dlDotDa/2 * aDot * (t - t0)^2
        l0      = getMeanLongitude(orbIni);
        aDot    = finalState.getOrbit().getADot().getAllDerivatives()[0];
        dlDotDa = finalState.getOrbit().getLMDot().getAllDerivatives()[1];
        final double dl = getMeanLongitude(finalState.toSpacecraftState().getOrbit()) - l0;
        a0Mas = dl / (dlDotDa * cycleDuration) - 0.5 * aDot * cycleDuration;
    }

    /**
     * Computes the total dV needed to perform the SK.
     * @param initialState  the state at the cycle start
     * @param cycleDuration the duration of the SK cycle
     * @return the total dV needed
     */
    private double totalDV(final SpacecraftState initialState,
                           final double cycleDuration) {

        // longitude excursion when the cycle is centered from tPeak - cycleDuration/2
        // to tPeak + cycleDuration/2 where tPeak is the date at which peak longitude
        // is achieved (i.e. dl/dt = 0)
        final double longitudeExcursion = -dlDotDa * aDot * cycleDuration * cycleDuration / 8;

        // we want to achieve a longitude at next cycle first maneuver which will be
        // exactly at the start of a centered parabola
        final double lEnd    = center - 0.5 * longitudeExcursion;

        // use the quadratic longitude model to compute the initial semi-major axis offset
        // needed to achieve a longitude at next cycle first maneuver which will be exactly
        // at the start of a centered parabola
        final double lEndMl0 = lEnd - l0;
        final double aMas    = lEndMl0 / (dlDotDa * cycleDuration) - 0.5 * aDot * cycleDuration;

        // compute the in-plane maneuver required to get this initial semi-major axis offset
        final double deltaA  = aMas - a0Mas;
        final double mu      = initialState.getMu();
        final double a       = initialState.getA();
        final double totalDV = thrustSignVelocity(initialState) *
                               FastMath.sqrt(mu * (2 / a - 1 / (a + deltaA))) - FastMath.sqrt(mu / a);
        return totalDV;
    }

    /**
     * Builds a fully configured {@link FieldDSSTPropagator}.
     * @param fieldOrbIni initial orbit
     * @return the FieldDSSTPropagator
     */
    private FieldDSSTPropagator<DerivativeStructure> createFieldPropagator(final FieldOrbit<DerivativeStructure> fieldOrbIni) {

        // General settings
        final double minStep = 1.0;
        final double maxStep = Constants.JULIAN_DAY;
        final double dP      = 1.0;
        final int    degree  = 6;
        final int    order   = 6;

        // Gets the field
        final Field<DerivativeStructure> field = fieldOrbIni.getA().getField();

        // Creates the integrator
        final double[][] tolerance = DSSTPropagator.tolerances(dP, fieldOrbIni.toOrbit());
        final AdaptiveStepsizeFieldIntegrator<DerivativeStructure> dp853 =
                        new DormandPrince853FieldIntegrator<>(field, minStep, maxStep, tolerance[0], tolerance[1]);
        dp853.setInitialStepSize(maxStep);

        // Creates the DSST propagator and add force models
        final FieldSpacecraftState<DerivativeStructure> state_0 = new FieldSpacecraftState<>(fieldOrbIni);
        final FieldDSSTPropagator<DerivativeStructure> field_propagator = new FieldDSSTPropagator<>(field, dp853);

        // Sets the initial state which is osculating here
        field_propagator.setInitialState(state_0, PropagationType.OSCULATING);

        // Earth potential
        final UnnormalizedSphericalHarmonicsProvider provider =
                GravityFieldFactory.getUnnormalizedProvider(degree, order);

        // Zonal harmonics
        final DSSTForceModel zonal = new DSSTZonal(provider, provider.getMaxDegree(),
                                                   4, 2 * provider.getMaxOrder() + 1);
        field_propagator.addForceModel(zonal);

        // Tesseral harmonics
        final DSSTForceModel tesseral = new DSSTTesseral(ecef,
                        Constants.WGS84_EARTH_ANGULAR_VELOCITY, provider,
                        provider.getMaxDegree(), provider.getMaxOrder(), 4, 12,
                        provider.getMaxDegree(), provider.getMaxOrder(), 4);
        field_propagator.addForceModel(tesseral);

        // Third bodies
        final DSSTForceModel nbody_sun = new DSSTThirdBody(CelestialBodyFactory.getSun(), MU);
        field_propagator.addForceModel(nbody_sun);
        final DSSTForceModel nbody_moon = new DSSTThirdBody(CelestialBodyFactory.getMoon(), MU);
        field_propagator.addForceModel(nbody_moon);

        return field_propagator;
    }

    /** Get the sign of the maneuver model with respect to orbital velocity.
     * @param state spacecraft state
     * @return +1 if model thrust is along velocity direction, -1 otherwise
     */
    private double thrustSignVelocity(final SpacecraftState state) {
        final Vector3D thrustDirection =
                state.getAttitude().getRotation().applyInverseTo(Vector3D.PLUS_I);
        final Vector3D velocity = state.getPVCoordinates().getVelocity();
        return FastMath.signum(Vector3D.dotProduct(thrustDirection, velocity));
    }

    /** Program entry point.
     * @param args program arguments (unused here)
     */
    public static void main(final String[] args) {

        try {

            // Configures access to Orekit data
            final File home = setupOrekit();

            // Reads input data from the input file
            final TutorialStationKeeping input = readInput();

            // Recovers the West and East boundaries of the longitude window
            final double westBound = FastMath.toRadians(Objects.requireNonNull(input.getWesternBoundary()));
            final double eastBound = FastMath.toRadians(Objects.requireNonNull(input.getEasternBoundary()));

            System.out.format("Window E/W: %f°, %f°", FastMath.toDegrees(eastBound),
                                                      FastMath.toDegrees(westBound));
            System.out.format(" - Center: %f°%n%n",
                              FastMath.toDegrees(0.5 * (westBound + MathUtils.normalizeAngle(eastBound, westBound))));

            // Recovers maneuvering capabilities
            final double dVMax = FastMath.abs(Objects.requireNonNull(input.getDeltaVMax()));
            final double isp   = Objects.requireNonNull(input.getIsp());

            // Recovers maneuvering constraints
            final int nbMaxMan = requireNonNullElse(input.getNbMaxManeuvers(), 1);
            final int orbitsNb = requireNonNullElse(input.getOrbitsSeparation(), 1);

            // Recovers data related to the SK cycle
            final double cycleSpan = Objects.requireNonNull(input.getCycleDuration()) * Constants.JULIAN_DAY;
            final int cyclesNb     = Objects.requireNonNull(input.getCyclesNumber());

            // Recovers the initial state for the simulation
            SpacecraftState initialState = getInitialState(input);
            final AbsoluteDate t0        = initialState.getDate();
            final double initialMass     = initialState.getMass();

            // Builds the SK calculator
            final StationKeeping sk = new StationKeeping(westBound, eastBound,
                                                         dVMax, isp, nbMaxMan, orbitsNb);

            // Gets data related to the SK cycle
            final boolean printToFile = requireNonNullElse(input.isPrintToFile(), false);

            // Create the numerical propagator
            final NumericalPropagator propagator = createPropagator(initialState);

            final SKHandler handler = new SKHandler();
            if (printToFile) {
                // Adds the step handler
                final double outputstep = requireNonNullElse(input.getOutputStep(),
                                                             Constants.JULIAN_DAY);
                propagator.getMultiplexer().add(outputstep, handler);
            }

            final long tc0 = System.currentTimeMillis();
            int manNb = 0;
            double dVtot = 0;
            // Loop over the cycles up to tEnd
            for (int i = 0; i < cyclesNb; i++) {
                System.out.format("> Cycle #%d from %s to %s %n",
                                  i + 1,
                                  format(initialState.getDate()),
                                  format(initialState.getDate().shiftedBy(cycleSpan)));

                final double l0 = sk.getMeanLongitude(initialState.getOrbit());
                System.out.format("   Longitude at start of cycle: %.3f°%n", FastMath.toDegrees(l0));

                // Computes the maneuvers for the current cycle
                final List<ImpulseManeuver> maneuvers = sk.run(initialState, cycleSpan);

                printManeuvers(maneuvers, t0);
                manNb += maneuvers.size();
                dVtot += maneuvers.stream().mapToDouble(m -> m.getDeltaVSat().getNorm()).sum();

                // Removes previous maneuvers (already triggered) from the propagator
                propagator.clearEventsDetectors();

                // Adds the new computed maneuvers to the propagator
                for (final ImpulseManeuver man: maneuvers) {
                    propagator.addEventDetector(man);
                }

                // Propagates for this cycle
                initialState = propagator.propagate(initialState.getDate().shiftedBy(cycleSpan));
            }

            System.out.println("\nBudget:");
            System.out.format("   Maneuvers nb : %d%n", manNb);
            System.out.format("   Total deltaV : %.3f m/s%n", dVtot);
            System.out.format("   Mass consumed: %.3f kg%n", initialMass - initialState.getMass());

            final long tc1 = System.currentTimeMillis();
            System.out.format("%nCalculation time: %.1f s%n", (tc1 - tc0) * 1.e-3);

            if (printToFile) {
                final File output = new File(home, "station-keeping.csv");
                try (PrintWriter csvOutput = new PrintWriter(output, StandardCharsets.UTF_8.toString())) {
                    // Print header
                    csvOutput.println("duration (d), sma (m), ex, ey, hx (°), hy (°), longitude (°)");
                    // Print orbital elements from collected states
                    for (final SpacecraftState state: handler.getStates()) {
                        csvOutput.println(state.getDate().durationFrom(t0) / Constants.JULIAN_DAY + SEP +
                                          state.getA() + SEP +
                                          state.getEquinoctialEx() + SEP +
                                          state.getEquinoctialEy() + SEP +
                                          FastMath.toDegrees(state.getHx()) + SEP +
                                          FastMath.toDegrees(state.getHy()) + SEP +
                                          FastMath.toDegrees(sk.getMeanLongitude(state.getOrbit())));
                    }
                    System.out.format(Locale.US, "%nOutput written to %s%n", output.getAbsolutePath());
                } catch (FileNotFoundException e) {
                    System.err.println(e.getLocalizedMessage());
                }
            }

            System.exit(0);

        } catch (IOException | URISyntaxException e) {
            System.err.println(e.getLocalizedMessage());
            System.exit(1);
        }
    }

    /**
     * Configure access to Orekit data and return home directory.
     * @return home directory
     */
    private static File setupOrekit() {
        // configure Orekit
        final File home       = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, unzip it in %s and rename it 'orekit-data' for this tutorial to work%n",
                              "orekit-data-master.zip", "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                              home.getAbsolutePath());
            System.exit(1);
        }
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));

        return home;
    }

    /**
     * Read the input data from the yaml input file.
     * @return the input data
     * @throws URISyntaxException if yaml input file can't be found
     * @throws IOException if yaml input file can't be read
     */
    private static TutorialStationKeeping readInput() throws URISyntaxException, IOException {
        // set input file
        final File input = new File(StationKeeping.class.getResource("/station-keeping.yaml").toURI().getPath());
        // read then return input data
        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        return mapper.readValue(input, TutorialStationKeeping.class);
    }

    /**
     * Gets the initial state from the input data.
     * @param input the input data
     * @return the initial state
     * @throws IOException if the orbit can't be built from the input data
     */
    private static SpacecraftState getInitialState(final TutorialStationKeeping input)
        throws IOException {

        // All dates are expressed in UTC time scale
        final TimeScale utc = TimeScalesFactory.getUTC();

        // Orbit data
        final TutorialOrbit orbitData = Objects.requireNonNull(input.getOrbit());

        final Frame frame;
        if (Objects.isNull(orbitData.getFrameName())) {
            frame = FramesFactory.getEME2000();
        } else {
            frame = orbitData.getInertialFrame();
        }

        // Orbit definition
        final AbsoluteDate orbitDate = new AbsoluteDate(Objects.requireNonNull(orbitData.getDate()), utc);
        final TutorialOrbitType orbitType = Objects.requireNonNull(orbitData.getOrbitType());
        final Orbit orbit;
        if (orbitType.getKeplerian() != null) {
            final TutorialKeplerianOrbit keplerian = orbitType.getKeplerian();
            orbit = new KeplerianOrbit(Objects.requireNonNull(keplerian.getA()),
                                       Objects.requireNonNull(keplerian.getE()),
                                       FastMath.toRadians(Objects.requireNonNull(keplerian.getI())),
                                       FastMath.toRadians(Objects.requireNonNull(keplerian.getPa())),
                                       FastMath.toRadians(Objects.requireNonNull(keplerian.getRaan())),
                                       FastMath.toRadians(Objects.requireNonNull(keplerian.getV())),
                                       PositionAngleType.valueOf(Objects.requireNonNull(keplerian.getPositionAngle())),
                                       frame, orbitDate, MU);
        } else if (orbitType.getEquinoctial() != null) {
            final TutorialEquinoctialOrbit equinoctial = orbitType.getEquinoctial();
            orbit = new EquinoctialOrbit(Objects.requireNonNull(equinoctial.getA()),
                                         Objects.requireNonNull(equinoctial.getEx()),
                                         Objects.requireNonNull(equinoctial.getEy()),
                                         Objects.requireNonNull(equinoctial.getHx()),
                                         Objects.requireNonNull(equinoctial.getHy()),
                                         FastMath.toRadians(Objects.requireNonNull(equinoctial.getLv())),
                                         PositionAngleType.valueOf(Objects.requireNonNull(equinoctial.getPositionAngle())),
                                         frame, orbitDate, MU);
        } else if (orbitType.getCircular() != null) {
            final TutorialCircularOrbit circular = orbitType.getCircular();
            orbit = new CircularOrbit(Objects.requireNonNull(circular.getA()),
                                      Objects.requireNonNull(circular.getEx()),
                                      Objects.requireNonNull(circular.getEy()),
                                      FastMath.toRadians(Objects.requireNonNull(circular.getI())),
                                      FastMath.toRadians(Objects.requireNonNull(circular.getRaan())),
                                      FastMath.toRadians(Objects.requireNonNull(circular.getAlphaV())),
                                      PositionAngleType.valueOf(Objects.requireNonNull(circular.getPositionAngle())),
                                      frame, orbitDate, MU);
        } else if (orbitType.getCartesian() != null) {
            final TutorialCartesianOrbit cartesian = orbitType.getCartesian();
            final double[] pos = {Objects.requireNonNull(cartesian.getX()),
                                  Objects.requireNonNull(cartesian.getY()),
                                  Objects.requireNonNull(cartesian.getZ())};
            final double[] vel = {Objects.requireNonNull(cartesian.getVx()),
                                  Objects.requireNonNull(cartesian.getVy()),
                                  Objects.requireNonNull(cartesian.getVz())};
            orbit = new CartesianOrbit(new PVCoordinates(new Vector3D(pos), new Vector3D(vel)),
                                       frame, orbitDate, MU);
        } else {
            throw new IOException("Unexpected orbit definition.");
        }

        // Gets initial mass
        final double mass = requireNonNullElse(input.getInitialMass(), 1000.);

        // Returns the spacecraft state whose attitude is aligned with VNC LOF.
        final Attitude attitude = new LofOffset(orbit.getFrame(), LOFType.VNC)
                                      .getAttitude(orbit, orbit.getDate(), orbit.getFrame());
        return new SpacecraftState(orbit, attitude, mass);

    }

    /** Builds a fully configured {@link NumericalPropagator}.
     * @param state the spacecraft state
     * @return the NumericalPropagator
     */
    private static NumericalPropagator createPropagator(final SpacecraftState state) {
        // general settings
        final double minStep = 0.1;
        final double maxStep = 300.;
        final double dP      = 1.0;
        final int    degree  = 10;
        final int    order   = 10;

        // create propagator
        final double[][] tol = NumericalPropagator.tolerances(dP, state.getOrbit(),
                                                                                state.getOrbit().getType());
        final AdaptiveStepsizeIntegrator dp853 = new DormandPrince853Integrator(minStep, maxStep, tol[0], tol[1]);
        dp853.setInitialStepSize(maxStep);
        final NumericalPropagator np = new NumericalPropagator(dp853);

        // set initial state
        np.resetInitialState(state);

        // general settings
        np.setAttitudeProvider(new LofOffset(state.getFrame(), LOFType.VNC));
        np.setOrbitType(state.getOrbit().getType());

        // Earth gravity field
        final NormalizedSphericalHarmonicsProvider gravityProvider = GravityFieldFactory.getNormalizedProvider(degree, order);
        np.addForceModel(new HolmesFeatherstoneAttractionModel(CelestialBodyFactory.getEarth().getBodyOrientedFrame(),
                                                               gravityProvider));

        // Sun and Moon attraction
        np.addForceModel(new ThirdBodyAttraction(CelestialBodyFactory.getSun()));
        np.addForceModel(new ThirdBodyAttraction(CelestialBodyFactory.getMoon()));

        return np;
    }

    /** Print the maneuvers.
     * @param maneuvers maneuvers list
     * @param tIni initial date
     */
    private static void printManeuvers(final List<ImpulseManeuver> maneuvers,
                                       final AbsoluteDate tIni) {
        int i = 0;
        for (ImpulseManeuver man: maneuvers) {
            final DateDetector manDateDetector = (DateDetector) man.getTrigger();
            final AbsoluteDate manDate = manDateDetector.getDates().get(0).getDate();
            final double dV = man.getDeltaVSat().getX();
            System.out.format("    DeltaV #%d = %f m/s @ %s (%.1f days from simulation start date)%n",
                              ++i,  dV, format(manDate),
                              manDate.durationFrom(tIni) / Constants.JULIAN_DAY);
        }
    }

    /**
     * Format a date (ISO8091, UTC scale, 3 digits for ms).
     * @param date the date to print
     * @return the date as a String in UTC scale with 3 digits for ms.
     */
    private static String format(final AbsoluteDate date) {
        return date.toStringWithoutUtcOffset(TimeScalesFactory.getUTC(), 3);
    }

    /**
     * Returns the first argument if it is non-{@code null} and
     * otherwise returns the non-{@code null} second argument.
     *
     * @param obj an object
     * @param defaultObj a non-{@code null} object to return if the first argument
     *                   is {@code null}
     * @param <T> the type of the reference
     * @return the first argument if it is non-{@code null} and
     *        otherwise the second argument if it is non-{@code null}
     * @throws NullPointerException if both {@code obj} is null and
     *        {@code defaultObj} is {@code null}
     */
    private static <T> T requireNonNullElse(final T obj, final T defaultObj) {
        return (obj != null) ? obj : Objects.requireNonNull(defaultObj, "defaultObj");
    }

    /** Step handler to be called at each step of the propagation. */
    private static class SKHandler implements OrekitFixedStepHandler {

        /** Sorted stored states. */
        private final SortedSet<SpacecraftState> states;

        /** Constructor. */
        SKHandler() {
            this.states = new TreeSet<>(new ChronologicalComparator());
        }

        /** Store the state at each sampling step.
         * @param state the state to store
         */
        public void handleStep(final SpacecraftState state) {
            states.add(state);
        }

        /** Get the stored states.
         * @return the stored states
         */
        public SortedSet<SpacecraftState> getStates() {
            return states;
        }
    }

    /**
     * Input data for this station-keeping tutorial.
     * <p>
     * Data are read from a YAML file.
     * </p>
     */
    public static class TutorialStationKeeping {

        /** Orbit data. */
        private TutorialOrbit orbit;

        /** Eastern boundary for longitude window (degrees in [-180, 180]). */
        private double easternBoundary;

        /** Western boundary for longitude window (degrees in [-180, 180]). */
        private double westernBoundary;

        /** DeltaV max (m/s). */
        private double deltaVMax;

        /** ISP (seconds). */
        private double isp;

        /** Initial mas (kg). */
        private double initialMass;

        /** Max number of maneuvers per cycle. */
        private int nbMaxManeuvers;

        /** Number of orbits between maneuvers. */
        private int orbitsSeparation;

        /** Cycle duration (days). */
        private double cycleDuration;

        /** Number of SK cycles to simulate. */
        private int cyclesNumber;

        /** Flag for printing results to a file. */
        private boolean printToFile;

        /** Time step between printed elements (s). */
        private double outputStep;

        /**
         * Get the orbit data.
         * @return the orbit data
         */
        public TutorialOrbit getOrbit() {
            return orbit;
        }

        /**
         * Set the orbit data.
         * @param orbit orbit data
         */
        public void setOrbit(final TutorialOrbit orbit) {
            this.orbit = orbit;
        }

        /**
         * Get the eastern boundary for longitude window.
         * @return the eastern boundary (degrees in [-180, 180])
         */
        public double getEasternBoundary() {
            return easternBoundary;
        }

        /**
         * Set the eastern boundary for longitude window.
         * @param easternBoundary the eastern boundary (degrees in [-180, 180])
         */
        public void setEasternBoundary(final double easternBoundary) {
            this.easternBoundary = easternBoundary;
        }

        /**
         * Get the western boundary for longitude window.
         * @return the western boundary (degrees in [-180, 180])
         */
        public double getWesternBoundary() {
            return westernBoundary;
        }

        /**
         * Set the western boundary for longitude window.
         * @param westernBoundary the western boundary (degrees in [-180, 180])
         */
        public void setWesternBoundary(final double westernBoundary) {
            this.westernBoundary = westernBoundary;
        }

        /**
         * Get the DV max.
         * @return the DV max (m/s)
         */
        public double getDeltaVMax() {
            return deltaVMax;
        }

        /**
         * Set the DV max.
         * @param deltaVMax DV max (m/s)
         */
        public void setDeltaVMax(final double deltaVMax) {
            this.deltaVMax = deltaVMax;
        }

        /**
         * Get the ISP.
         * @return the ISP (s)
         */
        public double getIsp() {
            return isp;
        }

        /**
         * Set the ISP.
         * @param isp ISP (s)
         */
        public void setIsp(final double isp) {
            this.isp = isp;
        }

        /**
         * Get the initial mass of the spacecraft.
         * @return the initial mass (kg)
         */
        public double getInitialMass() {
            return initialMass;
        }

        /**
         * Set the initial mass of the spacecraft.
         * @param initialMass the initial mass (kg)
         */
        public void setInitialMass(final double initialMass) {
            this.initialMass = initialMass;
        }

        /**
         * Get the max number of maneuvers per cycle.
         * @return the max number of maneuvers per cycle
         */
        public int getNbMaxManeuvers() {
            return nbMaxManeuvers;
        }

        /**
         * Set the max number of maneuvers per cycle.
         * @param nbMaxManeuvers the max number of maneuvers per cycle
         */
        public void setNbMaxManeuvers(final int nbMaxManeuvers) {
            this.nbMaxManeuvers = nbMaxManeuvers;
        }

        /**
         * Get the number of orbits between maneuvers.
         * @return the number of orbits between maneuvers
         */
        public int getOrbitsSeparation() {
            return orbitsSeparation;
        }

        /**
         * Set the number of orbits between maneuvers.
         * @param orbitsSeparation the number of orbits between maneuvers
         */
        public void setOrbitsSeparation(final int orbitsSeparation) {
            this.orbitsSeparation = orbitsSeparation;
        }

        /**
         * Get the SK cycle duration.
         * @return the cycle duration (days)
         */
        public double getCycleDuration() {
            return cycleDuration;
        }

        /**
         * Set the SK cycle duration.
         * @param cycleDuration cycle duration (days)
         */
        public void setCycleDuration(final double cycleDuration) {
            this.cycleDuration = cycleDuration;
        }

        /**
         * Get the number of SK cycles to simulate.
         * @return the number of SK cycles to simulate
         */
        public int getCyclesNumber() {
            return cyclesNumber;
        }

        /**
         * Set the number of SK cycles to simulate.
         * @param cyclesNumber the number of SK cycles to simulate
         */
        public void setCyclesNumber(final int cyclesNumber) {
            this.cyclesNumber = cyclesNumber;
        }

        /**
         * Get the flag for printing results to a file.
         * @return if true, results are printed to a file
         */
        public boolean isPrintToFile() {
            return printToFile;
        }

        /**
         * Set the flag for printing results to a file.
         * @param printToFile if true, results are are printed to a file
         */
        public void setPrintToFile(final boolean printToFile) {
            this.printToFile = printToFile;
        }

        /**
         * Get the time step between printed elements.
         * @return the time step between printed elements (s)
         */
        public double getOutputStep() {
            return outputStep;
        }

        /**
         * Set the time step between printed elements.
         * @param outputStep time step between printed elements (s)
         */
        public void setOutputStep(final double outputStep) {
            this.outputStep = outputStep;
        }

    }

}
